# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2011,2016 Red Hat, Inc.
#
# Authors:
# Thomas Woerner <twoerner@redhat.com>

DBUS_INTERFACE_VERSION = 1
DBUS_INTERFACE_REVISION = 15

DBUS_INTERFACE = "org.fedoraproject.FirewallD%d" % DBUS_INTERFACE_VERSION
DBUS_INTERFACE_ZONE = DBUS_INTERFACE + ".zone"
DBUS_INTERFACE_POLICY = DBUS_INTERFACE + ".policy"
DBUS_INTERFACE_DIRECT = DBUS_INTERFACE + ".direct"
DBUS_INTERFACE_POLICIES = DBUS_INTERFACE + ".policies"
DBUS_INTERFACE_IPSET = DBUS_INTERFACE + ".ipset"
DBUS_INTERFACE_CONFIG = DBUS_INTERFACE + ".config"
DBUS_INTERFACE_CONFIG_ZONE = DBUS_INTERFACE_CONFIG + ".zone"
DBUS_INTERFACE_CONFIG_POLICY = DBUS_INTERFACE_CONFIG + ".policy"
DBUS_INTERFACE_CONFIG_SERVICE = DBUS_INTERFACE_CONFIG + ".service"
DBUS_INTERFACE_CONFIG_ICMPTYPE = DBUS_INTERFACE_CONFIG + ".icmptype"
DBUS_INTERFACE_CONFIG_POLICIES = DBUS_INTERFACE_CONFIG + ".policies"
DBUS_INTERFACE_CONFIG_DIRECT = DBUS_INTERFACE_CONFIG + ".direct"
DBUS_INTERFACE_CONFIG_IPSET = DBUS_INTERFACE_CONFIG + ".ipset"
DBUS_INTERFACE_CONFIG_HELPER = DBUS_INTERFACE_CONFIG + ".helper"

DBUS_PATH = "/org/fedoraproject/FirewallD%d" % DBUS_INTERFACE_VERSION
DBUS_PATH_CONFIG = DBUS_PATH + "/config"
DBUS_PATH_CONFIG_ICMPTYPE = DBUS_PATH + "/config/icmptype"
DBUS_PATH_CONFIG_SERVICE = DBUS_PATH + "/config/service"
DBUS_PATH_CONFIG_ZONE = DBUS_PATH + "/config/zone"
DBUS_PATH_CONFIG_POLICY = DBUS_PATH + "/config/policy"
DBUS_PATH_CONFIG_IPSET = DBUS_PATH + "/config/ipset"
DBUS_PATH_CONFIG_HELPER = DBUS_PATH + "/config/helper"

# Polkit actions
_PK_ACTION = "org.fedoraproject.FirewallD%d" % DBUS_INTERFACE_VERSION
PK_ACTION_POLICIES = _PK_ACTION + ".policies"
PK_ACTION_POLICIES_INFO = PK_ACTION_POLICIES + ".info"
PK_ACTION_CONFIG = _PK_ACTION + ".config"
PK_ACTION_CONFIG_INFO = PK_ACTION_CONFIG + ".info"
PK_ACTION_DIRECT = _PK_ACTION + ".direct"
PK_ACTION_DIRECT_INFO = PK_ACTION_DIRECT + ".info"
PK_ACTION_INFO = _PK_ACTION + ".info"
PK_ACTION_ALL = _PK_ACTION + ".all"  # implies all other actions



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2007-2016 Red Hat, Inc.
#
# Authors:
# Thomas Woerner <twoerner@redhat.com>

# translation
import locale

try:
    locale.setlocale(locale.LC_ALL, "")
except locale.Error:
    import os

    os.environ["LC_ALL"] = "C"
    locale.setlocale(locale.LC_ALL, "")

DOMAIN = "firewalld"
import gettext

gettext.install(domain=DOMAIN)

from . import dbus  # noqa: F401

# configuration
DAEMON_NAME = "firewalld"
CONFIG_NAME = "firewall-config"
APPLET_NAME = "firewall-applet"
DATADIR = "/usr/share/" + DAEMON_NAME
CONFIG_GLADE_NAME = CONFIG_NAME + ".glade"
COPYRIGHT = "(C) 2010-2017 Red Hat, Inc."
VERSION = "2.1.1"
AUTHORS = [
    "Thomas Woerner <twoerner@redhat.com>",
    "Jiri Popelka <jpopelka@redhat.com>",
    "Eric Garver <e@erig.me>",
]
LICENSE = gettext.gettext(
    "This program is free software; you can redistribute it and/or modify "
    "it under the terms of the GNU General Public License as published by "
    "the Free Software Foundation; either version 2 of the License, or "
    "(at your option) any later version.\n"
    "\n"
    "This program is distributed in the hope that it will be useful, "
    "but WITHOUT ANY WARRANTY; without even the implied warranty of "
    "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the "
    "GNU General Public License for more details.\n"
    "\n"
    "You should have received a copy of the GNU General Public License "
    "along with this program.  If not, see <http://www.gnu.org/licenses/>."
)
WEBSITE = "http://www.firewalld.org"


def set_system_config_paths(path):
    global ETC_FIREWALLD, FIREWALLD_CONF, ETC_FIREWALLD_ZONES, ETC_FIREWALLD_SERVICES, ETC_FIREWALLD_ICMPTYPES, ETC_FIREWALLD_IPSETS, ETC_FIREWALLD_HELPERS, FIREWALLD_DIRECT, LOCKDOWN_WHITELIST, ETC_FIREWALLD_POLICIES
    ETC_FIREWALLD = path
    FIREWALLD_CONF = path + "/firewalld.conf"
    ETC_FIREWALLD_ZONES = path + "/zones"
    ETC_FIREWALLD_SERVICES = path + "/services"
    ETC_FIREWALLD_ICMPTYPES = path + "/icmptypes"
    ETC_FIREWALLD_IPSETS = path + "/ipsets"
    ETC_FIREWALLD_HELPERS = path + "/helpers"
    ETC_FIREWALLD_POLICIES = path + "/policies"
    FIREWALLD_DIRECT = path + "/direct.xml"
    LOCKDOWN_WHITELIST = path + "/lockdown-whitelist.xml"


set_system_config_paths("/etc/firewalld")


def set_default_config_paths(path):
    global USR_LIB_FIREWALLD, FIREWALLD_ZONES, FIREWALLD_SERVICES, FIREWALLD_ICMPTYPES, FIREWALLD_IPSETS, FIREWALLD_HELPERS, FIREWALLD_POLICIES
    USR_LIB_FIREWALLD = path
    FIREWALLD_ZONES = path + "/zones"
    FIREWALLD_SERVICES = path + "/services"
    FIREWALLD_ICMPTYPES = path + "/icmptypes"
    FIREWALLD_IPSETS = path + "/ipsets"
    FIREWALLD_HELPERS = path + "/helpers"
    FIREWALLD_POLICIES = path + "/policies"


set_default_config_paths("/usr/lib/firewalld")

FIREWALLD_LOGFILE = "/var/log/firewalld"

FIREWALLD_LOGTARGET = "mixed"

FIREWALLD_PIDFILE = "/var/run/firewalld.pid"

FIREWALLD_TEMPDIR = "/run/firewalld"

SYSCONFIGDIR = "/etc/conf.d"
IFCFGDIR = "/etc/sysconfig/network-scripts"

SYSCTL_CONFIG = "/etc/sysctl.conf"

# commands used by backends
COMMANDS = {
    "ipv4": "/usr/bin/iptables",
    "ipv4-restore": "/usr/bin/iptables-restore",
    "ipv6": "/usr/bin/ip6tables",
    "ipv6-restore": "/usr/bin/ip6tables-restore",
    "eb": "/bin/false",
    "eb-restore": "/bin/false",
    "ipset": "/bin/false",
    "modprobe": "/usr/bin/modprobe",
    "rmmod": "/usr/bin/rmmod",
}

LOG_DENIED_VALUES = ["all", "unicast", "broadcast", "multicast", "off"]
AUTOMATIC_HELPERS_VALUES = ["yes", "no", "system"]
FIREWALL_BACKEND_VALUES = ["nftables", "iptables"]

# fallbacks: will be overloaded by firewalld.conf
FALLBACK_ZONE = "public"
FALLBACK_MINIMAL_MARK = 100
FALLBACK_CLEANUP_ON_EXIT = True
FALLBACK_CLEANUP_MODULES_ON_EXIT = False
FALLBACK_LOCKDOWN = False
FALLBACK_IPV6_RPFILTER = True
FALLBACK_INDIVIDUAL_CALLS = False
FALLBACK_LOG_DENIED = "off"
FALLBACK_AUTOMATIC_HELPERS = "no"
FALLBACK_FIREWALL_BACKEND = "nftables"
FALLBACK_FLUSH_ALL_ON_RELOAD = True
FALLBACK_RELOAD_POLICY = "INPUT:DROP,FORWARD:DROP,OUTPUT:DROP"
FALLBACK_RFC3964_IPV4 = True
FALLBACK_ALLOW_ZONE_DRIFTING = False
FALLBACK_NFTABLES_FLOWTABLE = "off"
FALLBACK_NFTABLES_COUNTERS = False



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2011-2016 Red Hat, Inc.
#
# Authors:
# Thomas Woerner <twoerner@redhat.com>

"""Base firewall settings"""

DEFAULT_ZONE_TARGET = "{chain}_{zone}"
DEFAULT_POLICY_TARGET = "CONTINUE"
DEFAULT_POLICY_PRIORITY = -1
DEFAULT_ZONE_PRIORITY = 0

ZONE_TARGETS = ["ACCEPT", "%%REJECT%%", "DROP", DEFAULT_ZONE_TARGET, "default"]

POLICY_TARGETS = ["ACCEPT", "REJECT", "DROP", "CONTINUE"]

SHORTCUTS = {
    "PREROUTING": "PRE",
    "POSTROUTING": "POST",
    "INPUT": "IN",
    "FORWARD": "FWD",
    "OUTPUT": "OUT",
}

REJECT_TYPES = {
    "ipv4": [
        "icmp-host-prohibited",
        "host-prohib",
        "icmp-net-unreachable",
        "net-unreach",
        "icmp-host-unreachable",
        "host-unreach",
        "icmp-port-unreachable",
        "port-unreach",
        "icmp-proto-unreachable",
        "proto-unreach",
        "icmp-net-prohibited",
        "net-prohib",
        "tcp-reset",
        "tcp-rst",
        "icmp-admin-prohibited",
        "admin-prohib",
    ],
    "ipv6": [
        "icmp6-adm-prohibited",
        "adm-prohibited",
        "icmp6-no-route",
        "no-route",
        "icmp6-addr-unreachable",
        "addr-unreach",
        "icmp6-port-unreachable",
        "port-unreach",
        "tcp-reset",
    ],
}

# ipset types that can be used as a source in zones
# The match-set option will be src or src,src according to the
# dimension of the ipset.
SOURCE_IPSET_TYPES = [
    "hash:ip",
    "hash:ip,port",
    "hash:ip,mark",
    "hash:net",
    "hash:net,port",
    "hash:net,iface",
    "hash:mac",
]



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2018 Red Hat, Inc.
#
# Authors:
# Eric Garver <egarver@redhat.com>

import os

from firewall import config
from firewall.errors import FirewallError

from firewall.core.fw_config import FirewallConfig
from firewall.core.io.zone import zone_reader
from firewall.core.io.service import service_reader
from firewall.core.io.ipset import ipset_reader
from firewall.core.io.icmptype import icmptype_reader
from firewall.core.io.helper import helper_reader
from firewall.core.io.policy import policy_reader
from firewall.core.io.direct import Direct
from firewall.core.io.lockdown_whitelist import LockdownWhitelist
from firewall.core.io.firewalld_conf import firewalld_conf


def check_on_disk_config(fw):
    fw_config = FirewallConfig(fw)

    try:
        _firewalld_conf = firewalld_conf(config.FIREWALLD_CONF)
        _firewalld_conf.read()
    except FirewallError as error:
        raise FirewallError(error.code, "'%s': %s" % (config.FIREWALLD_CONF, error.msg))
    except IOError:
        # defaults will be filled
        pass
    except Exception as msg:
        raise Exception("'%s': %s" % (config.FIREWALLD_CONF, msg))
    fw_config.set_firewalld_conf(_firewalld_conf)

    readers = {
        "ipset": {
            "reader": ipset_reader,
            "add": fw_config.add_ipset,
            "dirs": [config.FIREWALLD_IPSETS, config.ETC_FIREWALLD_IPSETS],
        },
        "helper": {
            "reader": helper_reader,
            "add": fw_config.add_helper,
            "dirs": [config.FIREWALLD_HELPERS, config.ETC_FIREWALLD_HELPERS],
        },
        "icmptype": {
            "reader": icmptype_reader,
            "add": fw_config.add_icmptype,
            "dirs": [config.FIREWALLD_ICMPTYPES, config.ETC_FIREWALLD_ICMPTYPES],
        },
        "service": {
            "reader": service_reader,
            "add": fw_config.add_service,
            "dirs": [config.FIREWALLD_SERVICES, config.ETC_FIREWALLD_SERVICES],
        },
        "zone": {
            "reader": zone_reader,
            "add": fw_config.add_zone,
            "dirs": [config.FIREWALLD_ZONES, config.ETC_FIREWALLD_ZONES],
        },
        "policy": {
            "reader": policy_reader,
            "add": fw_config.add_policy_object,
            "dirs": [config.FIREWALLD_POLICIES, config.ETC_FIREWALLD_POLICIES],
        },
    }
    for reader in readers.keys():
        for _dir in readers[reader]["dirs"]:
            if not os.path.isdir(_dir):
                continue
            for file in sorted(os.listdir(_dir)):
                if file.endswith(".xml"):
                    obj = readers[reader]["reader"](file, _dir)
                    readers[reader]["add"](obj)
    fw_config.full_check_config()

    if os.path.isfile(config.FIREWALLD_DIRECT):
        try:
            obj = Direct(config.FIREWALLD_DIRECT)
            obj.read()
            obj.check_config(obj.export_config())
        except FirewallError as error:
            raise FirewallError(
                error.code, "'%s': %s" % (config.FIREWALLD_DIRECT, error.msg)
            )
        except Exception as msg:
            raise Exception("'%s': %s" % (config.FIREWALLD_DIRECT, msg))
    if os.path.isfile(config.LOCKDOWN_WHITELIST):
        try:
            obj = LockdownWhitelist(config.LOCKDOWN_WHITELIST)
            obj.read()
            obj.check_config(obj.export_config())
        except FirewallError as error:
            raise FirewallError(
                error.code, "'%s': %s" % (config.LOCKDOWN_WHITELIST, error.msg)
            )
        except Exception as msg:
            raise Exception("'%s': %s" % (config.LOCKDOWN_WHITELIST, msg))



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2010-2016 Red Hat, Inc.
#
# Authors:
# Thomas Woerner <twoerner@redhat.com>

import os
import sys
import copy
import time
import traceback
from typing import Dict, List
from firewall import config
from firewall import functions
from firewall.core import ipXtables
from firewall.core import ebtables
from firewall.core import nftables
from firewall.core import ipset
from firewall.core import modules
from firewall.core.fw_icmptype import FirewallIcmpType
from firewall.core.fw_service import FirewallService
from firewall.core.fw_zone import FirewallZone
from firewall.core.fw_direct import FirewallDirect
from firewall.core.fw_config import FirewallConfig
from firewall.core.fw_policies import FirewallPolicies
from firewall.core.fw_ipset import FirewallIPSet
from firewall.core.fw_transaction import FirewallTransaction
from firewall.core.fw_helper import FirewallHelper
from firewall.core.fw_policy import FirewallPolicy
from firewall.core.fw_nm import nm_get_bus_name, nm_get_interfaces_in_zone
from firewall.core.logger import log
from firewall.core.io.io_object import IO_Object
from firewall.core.io.firewalld_conf import firewalld_conf
from firewall.core.io.direct import Direct
from firewall.core.io.service import service_reader
from firewall.core.io.icmptype import icmptype_reader
from firewall.core.io.zone import zone_reader, Zone
from firewall.core.io.ipset import ipset_reader
from firewall.core.ipset import IPSET_TYPES
from firewall.core.io.helper import helper_reader
from firewall.core.io.policy import policy_reader
from firewall.core.io.functions import check_on_disk_config
from firewall.core.rich import Rich_Rule
from firewall import errors
from firewall.errors import FirewallError

############################################################################
#
# class Firewall
#
############################################################################


class Firewall:
    def __init__(self, offline=False):
        self._firewalld_conf = firewalld_conf(config.FIREWALLD_CONF)
        self._offline = offline

        if not offline:
            self.ip4tables_backend = ipXtables.ip4tables(self)
            self.ip6tables_backend = ipXtables.ip6tables(self)
            self.ebtables_backend = ebtables.ebtables()
            self.ipset_backend = ipset.ipset()
            self.nftables_backend = nftables.nftables(self)
            self.modules_backend = modules.modules()

        self.icmptype = FirewallIcmpType(self)
        self.service = FirewallService(self)
        self.zone = FirewallZone(self)
        self.direct = FirewallDirect(self)
        self.config = FirewallConfig(self)
        self.policies = FirewallPolicies()
        self.ipset = FirewallIPSet(self)
        self.helper = FirewallHelper(self)
        self.policy = FirewallPolicy(self)

        self.__init_vars()

    def __repr__(self):
        return "%s(%r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r)" % (
            self.__class__,
            self.ip4tables_enabled,
            self.ip6tables_enabled,
            self.ebtables_enabled,
            self._state,
            self._panic,
            self._default_zone,
            self._module_refcount,
            self._marks,
            self.cleanup_on_exit,
            self.cleanup_modules_on_exit,
            self.ipv6_rpfilter_enabled,
            self.ipset_enabled,
            self._individual_calls,
            self._log_denied,
        )

    def __init_vars(self):
        self._state = "INIT"
        self._panic = False
        self._default_zone = config.FALLBACK_ZONE
        self._default_zone_interfaces = []
        self._nm_assigned_interfaces = []
        self._module_refcount = {}
        self._marks = []
        # fallback settings will be overloaded by firewalld.conf
        self.cleanup_on_exit = config.FALLBACK_CLEANUP_ON_EXIT
        self.cleanup_modules_on_exit = config.FALLBACK_CLEANUP_MODULES_ON_EXIT
        self.ipv6_rpfilter_enabled = config.FALLBACK_IPV6_RPFILTER
        self._individual_calls = config.FALLBACK_INDIVIDUAL_CALLS
        self._log_denied = config.FALLBACK_LOG_DENIED
        self._firewall_backend = config.FALLBACK_FIREWALL_BACKEND
        self._flush_all_on_reload = config.FALLBACK_FLUSH_ALL_ON_RELOAD
        self._rfc3964_ipv4 = config.FALLBACK_RFC3964_IPV4
        self._allow_zone_drifting = config.FALLBACK_ALLOW_ZONE_DRIFTING
        self._nftables_flowtable = config.FALLBACK_NFTABLES_FLOWTABLE
        self._nftables_counters = config.FALLBACK_NFTABLES_COUNTERS

        if self._offline:
            self.ip4tables_enabled = False
            self.ip6tables_enabled = False
            self.ebtables_enabled = False
            self.ipset_enabled = False
            self.ipset_supported_types = IPSET_TYPES
            self.nftables_enabled = False
        else:
            self.ip4tables_enabled = True
            self.ipv4_supported_icmp_types = []
            self.ip6tables_enabled = True
            self.ipv6_supported_icmp_types = []
            self.ebtables_enabled = True
            self.ipset_enabled = True
            self.ipset_supported_types = IPSET_TYPES
            self.nftables_enabled = True

    def get_all_io_objects_dict(self):
        """
        Returns a dict of dicts of all runtime config objects.
        """
        conf_dict = {}
        conf_dict["ipsets"] = {
            _ipset: self.ipset.get_ipset(_ipset) for _ipset in self.ipset.get_ipsets()
        }
        conf_dict["helpers"] = {
            helper: self.helper.get_helper(helper)
            for helper in self.helper.get_helpers()
        }
        conf_dict["icmptypes"] = {
            icmptype: self.icmptype.get_icmptype(icmptype)
            for icmptype in self.icmptype.get_icmptypes()
        }
        conf_dict["services"] = {
            service: self.service.get_service(service)
            for service in self.service.get_services()
        }
        conf_dict["zones"] = {
            zone: self.zone.get_zone(zone) for zone in self.zone.get_zones()
        }
        conf_dict["policies"] = {
            policy: self.policy.get_policy(policy)
            for policy in self.policy.get_policies_not_derived_from_zone()
        }

        conf_dict["conf"] = {}
        conf_dict["conf"]["FirewallBackend"] = self._firewalld_conf.get(
            "FirewallBackend"
        )

        # The runtime might not actually support all the defined icmptypes.
        # This is the case if ipv6 (ip6tables) is disabled. Unfortunately users
        # disable IPv6 and also expect the IPv6 stuff to be silently ignored.
        # This is problematic for defaults that include IPv6 stuff, e.g. policy
        # 'allow-host-ipv6'. Use this to make a better decision about errors vs
        # warnings.
        #
        conf_dict["runtime"] = {}
        conf_dict["runtime"]["icmptypes_unsupported"] = {}
        for icmptype in set(self.config.get_icmptypes()).difference(
            set(self.icmptype.get_icmptypes())
        ):
            conf_dict["runtime"]["icmptypes_unsupported"][
                icmptype
            ] = self.config.get_icmptype(icmptype)
        # Some icmptypes support multiple families. Add those that are missing
        # support for a subset of families.
        for icmptype in set(self.config.get_icmptypes()).intersection(
            set(self.icmptype.get_icmptypes())
        ):
            if (
                icmptype not in self.ipv4_supported_icmp_types
                or icmptype not in self.ipv6_supported_icmp_types
            ):
                conf_dict["runtime"]["icmptypes_unsupported"][icmptype] = copy.copy(
                    self.config.get_icmptype(icmptype)
                )
                conf_dict["runtime"]["icmptypes_unsupported"][icmptype].destination = []
                if icmptype not in self.ipv4_supported_icmp_types:
                    conf_dict["runtime"]["icmptypes_unsupported"][
                        icmptype
                    ].destination.append("ipv4")
                if icmptype not in self.ipv6_supported_icmp_types:
                    conf_dict["runtime"]["icmptypes_unsupported"][
                        icmptype
                    ].destination.append("ipv6")

        return conf_dict

    def full_check_config(self, extra_io_objects: Dict[str, List[IO_Object]] = {}):
        all_io_objects = self.get_all_io_objects_dict()
        # mix in the extra objects
        for type_key in extra_io_objects:
            for obj in extra_io_objects[type_key]:
                all_io_objects[type_key][obj.name] = obj

        # we need to check in a well defined order because some io_objects will
        # cross-check others
        order = ["ipsets", "helpers", "icmptypes", "services", "zones", "policies"]
        for io_obj_type in order:
            io_objs = all_io_objects[io_obj_type]
            for name, io_obj in io_objs.items():
                io_obj.check_config_dict(io_obj.export_config_dict(), all_io_objects)

    def _start_check_tables(self):
        # check if iptables, ip6tables and ebtables are usable, else disable
        if (
            self.ip4tables_enabled
            and "filter" not in self.ip4tables_backend.get_available_tables()
        ):
            log.info1("iptables is not usable.")
            self.ip4tables_enabled = False

        if (
            self.ip6tables_enabled
            and "filter" not in self.ip6tables_backend.get_available_tables()
        ):
            log.info1("ip6tables is not usable.")
            self.ip6tables_enabled = False

        if (
            self.ebtables_enabled
            and "filter" not in self.ebtables_backend.get_available_tables()
        ):
            log.info1("ebtables is not usable.")
            self.ebtables_enabled = False

        # is there at least support for ipv4 or ipv6
        if (
            not self.ip4tables_enabled
            and not self.ip6tables_enabled
            and not self.nftables_enabled
        ):
            raise FirewallError(errors.UNKNOWN_ERROR, "No IPv4 and IPv6 firewall.")

    def _start_probe_backends(self):
        try:
            self.ipset_backend.set_list()
        except ValueError:
            if self.nftables_enabled:
                log.info1(
                    "ipset not usable, disabling ipset usage in firewall. Other set backends (nftables) remain usable."
                )
            else:
                log.warning("ipset not usable, disabling ipset usage in firewall.")
                self.ipset_supported_types = []
            # ipset is not usable
            self.ipset_enabled = False
        else:
            # ipset is usable, get all supported types
            self.ipset_supported_types = self.ipset_backend.set_supported_types()

        self.ip4tables_backend.fill_exists()
        if not self.ip4tables_backend.restore_command_exists:
            if self.ip4tables_backend.command_exists:
                log.warning(
                    "iptables-restore is missing, using "
                    "individual calls for IPv4 firewall."
                )
            else:
                if self.nftables_enabled:
                    log.info1(
                        "iptables-restore and iptables are missing, "
                        "IPv4 direct rules won't be usable."
                    )
                else:
                    log.warning(
                        "iptables-restore and iptables are missing, "
                        "disabling IPv4 firewall."
                    )
                self.ip4tables_enabled = False
        if self.nftables_enabled:
            self.ipv4_supported_icmp_types = self.nftables_backend.supported_icmp_types(
                "ipv4"
            )
        else:
            if self.ip4tables_enabled:
                self.ipv4_supported_icmp_types = (
                    self.ip4tables_backend.supported_icmp_types()
                )
            else:
                self.ipv4_supported_icmp_types = []
        self.ip6tables_backend.fill_exists()
        if not self.ip6tables_backend.restore_command_exists:
            if self.ip6tables_backend.command_exists:
                log.warning(
                    "ip6tables-restore is missing, using "
                    "individual calls for IPv6 firewall."
                )
            else:
                if self.nftables_enabled:
                    log.info1(
                        "ip6tables-restore and ip6tables are missing, "
                        "IPv6 direct rules won't be usable."
                    )
                else:
                    log.warning(
                        "ip6tables-restore and ip6tables are missing, "
                        "disabling IPv6 firewall."
                    )
                self.ip6tables_enabled = False
        if self.nftables_enabled:
            self.ipv6_supported_icmp_types = self.nftables_backend.supported_icmp_types(
                "ipv6"
            )
        else:
            if self.ip6tables_enabled:
                self.ipv6_supported_icmp_types = (
                    self.ip6tables_backend.supported_icmp_types()
                )
            else:
                self.ipv6_supported_icmp_types = []
        self.ebtables_backend.fill_exists()
        if not self.ebtables_backend.restore_command_exists:
            if self.ebtables_backend.command_exists:
                log.warning(
                    "ebtables-restore is missing, using "
                    "individual calls for bridge firewall."
                )
            else:
                if self.nftables_enabled:
                    log.info1(
                        "ebtables-restore and ebtables are missing, "
                        "eb direct rules won't be usable."
                    )
                else:
                    log.warning(
                        "ebtables-restore and ebtables are missing, "
                        "disabling bridge firewall."
                    )
                self.ebtables_enabled = False

        if (
            self.ebtables_enabled
            and not self._individual_calls
            and not self.ebtables_backend.restore_noflush_option
        ):
            log.debug1(
                "ebtables-restore is not supporting the --noflush "
                "option, will therefore not be used"
            )

    def _start_load_firewalld_conf(self):
        # load firewalld config
        log.debug1("Loading firewalld config file '%s'", config.FIREWALLD_CONF)
        try:
            self._firewalld_conf.read()
        except Exception as msg:
            log.warning(msg)
            log.warning("Using fallback firewalld configuration settings.")
        else:
            if self._firewalld_conf.get("DefaultZone"):
                self._default_zone = self._firewalld_conf.get("DefaultZone")

            if self._firewalld_conf.get("CleanupOnExit"):
                value = self._firewalld_conf.get("CleanupOnExit")
                if value is not None and value.lower() in ["no", "false"]:
                    self.cleanup_on_exit = False
                log.debug1("CleanupOnExit is set to '%s'", self.cleanup_on_exit)

            if self._firewalld_conf.get("CleanupModulesOnExit"):
                value = self._firewalld_conf.get("CleanupModulesOnExit")
                if value is not None and value.lower() in ["yes", "true"]:
                    self.cleanup_modules_on_exit = True
                log.debug1(
                    "CleanupModulesOnExit is set to '%s'", self.cleanup_modules_on_exit
                )

            if self._firewalld_conf.get("Lockdown"):
                value = self._firewalld_conf.get("Lockdown")
                if value is not None and value.lower() in ["yes", "true"]:
                    log.debug1("Lockdown is enabled")
                    try:
                        self.policies.enable_lockdown()
                    except FirewallError:
                        # already enabled, this is probably reload
                        pass

            if self._firewalld_conf.get("IPv6_rpfilter"):
                value = self._firewalld_conf.get("IPv6_rpfilter")
                if value is not None:
                    if value.lower() in ["no", "false"]:
                        self.ipv6_rpfilter_enabled = False
                    if value.lower() in ["yes", "true"]:
                        self.ipv6_rpfilter_enabled = True
            if self.ipv6_rpfilter_enabled:
                log.debug1("IPv6 rpfilter is enabled")
            else:
                log.debug1("IPV6 rpfilter is disabled")

            if self._firewalld_conf.get("IndividualCalls"):
                value = self._firewalld_conf.get("IndividualCalls")
                if value is not None and value.lower() in ["yes", "true"]:
                    log.debug1("IndividualCalls is enabled")
                    self._individual_calls = True

            if self._firewalld_conf.get("LogDenied"):
                value = self._firewalld_conf.get("LogDenied")
                if value is None or value.lower() == "no":
                    self._log_denied = "off"
                else:
                    self._log_denied = value.lower()
                    log.debug1("LogDenied is set to '%s'", self._log_denied)

            if self._firewalld_conf.get("FirewallBackend"):
                self._firewall_backend = self._firewalld_conf.get("FirewallBackend")
                log.debug1("FirewallBackend is set to '%s'", self._firewall_backend)

            if self._firewalld_conf.get("FlushAllOnReload"):
                value = self._firewalld_conf.get("FlushAllOnReload")
                if value.lower() in ["no", "false"]:
                    self._flush_all_on_reload = False
                else:
                    self._flush_all_on_reload = True
                log.debug1("FlushAllOnReload is set to '%s'", self._flush_all_on_reload)

            if self._firewalld_conf.get("RFC3964_IPv4"):
                value = self._firewalld_conf.get("RFC3964_IPv4")
                if value.lower() in ["no", "false"]:
                    self._rfc3964_ipv4 = False
                else:
                    self._rfc3964_ipv4 = True
                log.debug1("RFC3964_IPv4 is set to '%s'", self._rfc3964_ipv4)

            if self._firewalld_conf.get("NftablesFlowtable"):
                self._nftables_flowtable = self._firewalld_conf.get("NftablesFlowtable")
                log.debug1("NftablesFlowtable is set to '%s'", self._nftables_flowtable)

            if self._firewalld_conf.get("NftablesCounters"):
                value = self._firewalld_conf.get("NftablesCounters")
                if value.lower() in ["no", "false"]:
                    self._nftables_counters = False
                else:
                    self._nftables_counters = True
                log.debug1("NftablesCounters is set to '%s'", self._nftables_counters)

        self.config.set_firewalld_conf(copy.deepcopy(self._firewalld_conf))

    def _start_load_lockdown_whitelist(self):
        # load lockdown whitelist
        log.debug1("Loading lockdown whitelist")
        try:
            self.policies.lockdown_whitelist.read()
        except Exception as msg:
            if self.policies.query_lockdown():
                log.error(
                    "Failed to load lockdown whitelist '%s': %s",
                    self.policies.lockdown_whitelist.filename,
                    msg,
                )
            else:
                log.debug1(
                    "Failed to load lockdown whitelist '%s': %s",
                    self.policies.lockdown_whitelist.filename,
                    msg,
                )

        # copy policies to config interface
        self.config.set_policies(copy.deepcopy(self.policies))

    def _start_load_stock_config(self):
        self._loader_ipsets(config.FIREWALLD_IPSETS)
        self._loader_icmptypes(config.FIREWALLD_ICMPTYPES)
        self._loader_helpers(config.FIREWALLD_HELPERS)
        self._loader_services(config.FIREWALLD_SERVICES)
        self._loader_zones(config.FIREWALLD_ZONES)
        self._loader_policies(config.FIREWALLD_POLICIES)

    def _start_load_user_config(self):
        self._loader_ipsets(config.ETC_FIREWALLD_IPSETS)
        self._loader_icmptypes(config.ETC_FIREWALLD_ICMPTYPES)
        self._loader_helpers(config.ETC_FIREWALLD_HELPERS)
        self._loader_services(config.ETC_FIREWALLD_SERVICES)
        self._loader_zones(config.ETC_FIREWALLD_ZONES)
        self._loader_policies(config.ETC_FIREWALLD_POLICIES)

    def _start_copy_config_to_runtime(self):
        for _ipset in self.config.get_ipsets():
            self.ipset.add_ipset(copy.deepcopy(self.config.get_ipset(_ipset)))
        for icmptype in self.config.get_icmptypes():
            self.icmptype.add_icmptype(
                copy.deepcopy(self.config.get_icmptype(icmptype))
            )
        for helper in self.config.get_helpers():
            self.helper.add_helper(copy.deepcopy(self.config.get_helper(helper)))
        for service in self.config.get_services():
            self.service.add_service(copy.deepcopy(self.config.get_service(service)))
        for policy in self.config.get_policy_objects():
            self.policy.add_policy(copy.deepcopy(self.config.get_policy_object(policy)))

        self.direct.set_permanent_config(copy.deepcopy(self.config.get_direct()))

        # copy combined permanent zones to runtime
        # zones with a '/' in the name will be combined into one runtime zone
        combined_zones = {}
        for zone in self.config.get_zones():
            z_obj = self.config.get_zone(zone)
            if "/" not in z_obj.name:
                self.zone.add_zone(copy.deepcopy(self.config.get_zone(zone)))
                continue

            combined_name = os.path.basename(z_obj.path)
            if combined_name not in combined_zones:
                combined_zone = Zone()
                combined_zone.name = combined_name
                combined_zone.check_name(combined_zone.name)
                combined_zone.path = z_obj.path
                combined_zone.default = False
                combined_zone.forward = False  # see note in zone_reader()

                combined_zones[combined_name] = combined_zone

            log.debug1(
                "Combining zone '%s' using '%s%s%s'",
                combined_name,
                z_obj.path,
                os.sep,
                z_obj.filename,
            )
            combined_zones[combined_name].combine(z_obj)

        for zone in combined_zones:
            self.zone.add_zone(combined_zones[zone])

    def _start_load_direct_rules(self):
        # load direct rules
        obj = Direct(config.FIREWALLD_DIRECT)
        if os.path.exists(config.FIREWALLD_DIRECT):
            log.debug1("Loading direct rules file '%s'" % config.FIREWALLD_DIRECT)
            try:
                obj.read()
            except Exception as msg:
                log.error(
                    "Failed to load direct rules file '%s': %s",
                    config.FIREWALLD_DIRECT,
                    msg,
                )
        self.config.set_direct(obj)

    def _start_apply_objects(self, reload=False, complete_reload=False):
        transaction = FirewallTransaction(self)

        if not reload:
            self.flush(use_transaction=transaction)

        # If modules need to be unloaded in complete reload or if there are
        # ipsets to get applied, limit the transaction to flush.
        #
        # Future optimization for the ipset case in reload: The transaction
        # only needs to be split here if there are conflicting ipset types in
        # exsting ipsets and the configuration in firewalld.
        if (reload and complete_reload) or (
            self.ipset.backends() and self.ipset.has_ipsets()
        ):
            transaction.execute(True)
            transaction.clear()

        # complete reload: unload modules also
        if reload and complete_reload:
            log.debug1("Unloading firewall modules")
            self.modules_backend.unload_firewall_modules()

        self.apply_default_tables(use_transaction=transaction)
        transaction.execute(True)
        transaction.clear()

        # apply settings for loaded ipsets while reloading here
        if (self.ipset.backends()) and self.ipset.has_ipsets():
            log.debug1("Applying ipsets")
            self.ipset.apply_ipsets()

        log.debug1("Applying default rule set")
        self.apply_default_rules(use_transaction=transaction)

        log.debug1("Applying default zone")
        self.zone.apply_zone_settings(self._default_zone, transaction)
        self.zone._interface(True, self._default_zone, "+", transaction)

        log.debug1("Applying used zones")
        self.zone.apply_zones(use_transaction=transaction)

        log.debug1("Applying used policies")
        self.policy.apply_policies(use_transaction=transaction)

        transaction.execute(True)
        transaction.clear()

    def _start_apply_direct_rules(self):
        transaction = FirewallTransaction(self)

        # apply direct chains, rules and passthrough rules
        if self.direct.has_configuration():
            log.debug1("Applying direct chains rules and passthrough rules")
            self.direct.apply_direct(transaction)

            # since direct rules are easy to make syntax errors lets highlight
            # the cause if the transaction fails.
            try:
                transaction.execute(True)
                transaction.clear()
            except FirewallError as e:
                raise FirewallError(e.code, "Direct: %s" % (e.msg if e.msg else ""))
            except Exception:
                raise

        transaction.execute(True)
        transaction.clear()

    def _start_check(self):
        # check minimum required zones
        for z in ["block", "drop", "trusted"]:
            if z not in self.zone.get_zones():
                raise FirewallError(
                    errors.INVALID_ZONE, "Zone '{}' is not available.".format(z)
                )

        # check if default_zone is a valid zone
        if self._default_zone not in self.zone.get_zones():
            if "public" in self.zone.get_zones():
                zone = "public"
            elif "external" in self.zone.get_zones():
                zone = "external"
            else:
                zone = "block"  # block is a base zone, therefore it has to exist

            log.error(
                "Default zone '%s' is not valid. Using '%s'.", self._default_zone, zone
            )
            self._default_zone = zone
        else:
            log.debug1("Using default zone '%s'", self._default_zone)

        if not self._offline:
            self.full_check_config()
            self._start_check_tables()

            # check our desired backend is actually available
            if self._firewall_backend == "iptables":
                backend_to_check = "ip4tables"  # ip6tables is always optional
            else:
                backend_to_check = self._firewall_backend
            if not self.is_backend_enabled(backend_to_check):
                raise FirewallError(
                    errors.UNKNOWN_ERROR,
                    "Firewall backend '{}' is not available.".format(
                        self._firewall_backend
                    ),
                )

    def _start(self, reload=False, complete_reload=False):
        self._start_load_firewalld_conf()
        self._start_load_lockdown_whitelist()

        self._select_firewall_backend(self._firewall_backend)

        if not self._offline:
            self._start_probe_backends()

        self._start_load_stock_config()
        self._start_load_user_config()
        self._start_load_direct_rules()
        self._start_copy_config_to_runtime()

        self._start_check()

        if self._offline:
            return

        if log.getDebugLogLevel() > 0:
            # get time before flushing and applying
            tm1 = time.time()

        self._start_apply_objects(reload=reload, complete_reload=complete_reload)
        self._start_apply_direct_rules()

        if log.getDebugLogLevel() > 1:
            # get time after flushing and applying
            tm2 = time.time()
            log.debug2("Flushing and applying took %f seconds" % (tm2 - tm1))

    def _start_failsafe(self, reload=False, complete_reload=False):
        """
        This is basically _start() with at least the following differences:
            - built-in defaults for firewalld.conf
            - no lockdown list
            - no user config (/etc/firewalld)
            - no direct rules
        """
        self.cleanup()
        self._firewalld_conf.set_defaults()
        self.config.set_firewalld_conf(copy.deepcopy(self._firewalld_conf))

        self._select_firewall_backend(self._firewall_backend)

        if not self._offline:
            self._start_probe_backends()

        self._start_load_stock_config()
        self._start_copy_config_to_runtime()
        self._start_check()

        if self._offline:
            return

        self._start_apply_objects(reload=reload, complete_reload=complete_reload)

    def start(self):
        try:
            self._start()
        except Exception as original_ex:
            log.error(
                "Failed to load user configuration. Falling back to "
                "full stock configuration."
            )
            try:
                self._start_failsafe()
                self._state = "FAILED"
                self.set_policy("ACCEPT")
            except Exception as new_ex:
                log.error(original_ex)
                log.exception()
                log.error(new_ex)
                log.error(
                    "Failed to load full stock configuration. This likely "
                    "indicates a system level issue, e.g. the firewall "
                    "backend (nftables, iptables) is broken. "
                    "All hope is lost. Exiting."
                )
                try:
                    self.flush()
                except Exception:
                    pass
                sys.exit(errors.UNKNOWN_ERROR)
            # propagate the original exception that caused us to enter failed
            # state.
            raise original_ex
        else:
            self._state = "RUNNING"
            self.set_policy("ACCEPT")

    def _loader_config_file_generator(self, path):
        if not os.path.isdir(path):
            return

        for filename in sorted(os.listdir(path)):
            if not filename.endswith(".xml"):
                continue
            yield filename

    def _loader_services(self, path):
        for filename in self._loader_config_file_generator(path):
            log.debug1("Loading service file '%s%s%s'", path, os.sep, filename)

            obj = service_reader(filename, path)
            if obj.name in self.config.get_services():
                orig_obj = self.config.get_service(obj.name)
                log.debug1(
                    "Overrides '%s%s%s'", orig_obj.path, os.sep, orig_obj.filename
                )
            elif obj.path.startswith(config.ETC_FIREWALLD):
                obj.default = True

            self.config.add_service(obj)

    def _loader_ipsets(self, path):
        for filename in self._loader_config_file_generator(path):
            log.debug1("Loading ipset file '%s%s%s'", path, os.sep, filename)

            obj = ipset_reader(filename, path)
            if obj.name in self.config.get_ipsets():
                orig_obj = self.config.get_ipset(obj.name)
                log.debug1(
                    "Overrides '%s%s%s'", orig_obj.path, os.sep, orig_obj.filename
                )
            elif obj.path.startswith(config.ETC_FIREWALLD):
                obj.default = True

            self.config.add_ipset(obj)

    def _loader_helpers(self, path):
        for filename in self._loader_config_file_generator(path):
            log.debug1("Loading helper file '%s%s%s'", path, os.sep, filename)

            obj = helper_reader(filename, path)
            if obj.name in self.config.get_helpers():
                orig_obj = self.config.get_helper(obj.name)
                log.debug1(
                    "Overrides '%s%s%s'", orig_obj.path, os.sep, orig_obj.filename
                )
            elif obj.path.startswith(config.ETC_FIREWALLD):
                obj.default = True

            self.config.add_helper(obj)

    def _loader_policies(self, path):
        for filename in self._loader_config_file_generator(path):
            log.debug1("Loading policy file '%s%s%s'", path, os.sep, filename)

            obj = policy_reader(filename, path)
            if obj.name in self.config.get_policy_objects():
                orig_obj = self.config.get_policy_object(obj.name)
                log.debug1(
                    "Overrides '%s%s%s'", orig_obj.path, os.sep, orig_obj.filename
                )
            elif obj.path.startswith(config.ETC_FIREWALLD):
                obj.default = True

            self.config.add_policy_object(obj)

    def _loader_icmptypes(self, path):
        for filename in self._loader_config_file_generator(path):
            log.debug1("Loading icmptype file '%s%s%s'", path, os.sep, filename)

            obj = icmptype_reader(filename, path)
            if obj.name in self.config.get_icmptypes():
                orig_obj = self.config.get_icmptype(obj.name)
                log.debug1(
                    "Overrides '%s%s%s'", orig_obj.path, os.sep, orig_obj.filename
                )
            elif obj.path.startswith(config.ETC_FIREWALLD):
                obj.default = True

            self.config.add_icmptype(obj)

    def _loader_zones(self, path, combine=False):
        if not os.path.isdir(path):
            return

        for filename in sorted(os.listdir(path)):
            if not filename.endswith(".xml"):
                if path.startswith(config.ETC_FIREWALLD) and os.path.isdir(
                    "%s/%s" % (path, filename)
                ):
                    # Combined zones are added to permanent config
                    # individually. They're coalesced into one object when
                    # added to the runtime
                    self._loader_zones("%s/%s" % (path, filename), combine=True)
                continue

            name = "%s/%s" % (path, filename)
            log.debug1("Loading zone file '%s'", name)

            obj = zone_reader(filename, path, no_check_name=combine)
            if combine:
                # Change name for permanent configuration
                obj.name = "%s/%s" % (
                    os.path.basename(path),
                    os.path.basename(filename)[0:-4],
                )
                obj.check_name(obj.name)

            if obj.name in self.config.get_zones():
                orig_obj = self.config.get_zone(obj.name)
                log.debug1(
                    "Overrides '%s%s%s'", orig_obj.path, os.sep, orig_obj.filename
                )
            elif obj.path.startswith(config.ETC_FIREWALLD):
                obj.default = True

            self.config.add_zone(obj)

    def cleanup(self):
        self.icmptype.cleanup()
        self.service.cleanup()
        self.zone.cleanup()
        self.ipset.cleanup()
        self.helper.cleanup()
        self.config.cleanup()
        self.direct.cleanup()
        self.policies.cleanup()
        self.policy.cleanup()
        self._firewalld_conf.cleanup()
        self.__init_vars()

    def stop(self):
        if not self._offline:
            if self.cleanup_on_exit:
                self.flush()
                self.ipset.flush()
                self.set_policy("ACCEPT")

            if self.cleanup_modules_on_exit:
                log.debug1("Unloading firewall kernel modules")
                self.modules_backend.unload_firewall_modules()

        self.cleanup()

    # handle modules

    def handle_modules(self, _modules, enable):
        num_failed = 0
        error_msgs = ""
        for i, module in enumerate(_modules):
            if enable:
                (status, msg) = self.modules_backend.load_module(module)
            else:
                if self._module_refcount[module] > 1:
                    status = 0  # module referenced more then one, do not unload
                else:
                    (status, msg) = self.modules_backend.unload_module(module)
            if status != 0:
                num_failed += 1
                error_msgs += msg
                continue

            if enable:
                self._module_refcount.setdefault(module, 0)
                self._module_refcount[module] += 1
            else:
                if module in self._module_refcount:
                    self._module_refcount[module] -= 1
                    if self._module_refcount[module] == 0:
                        del self._module_refcount[module]
        return (num_failed, error_msgs)

    def _select_firewall_backend(self, backend):
        if backend != "nftables":
            self.nftables_enabled = False
        # even if using nftables, the other backends are enabled for use with
        # the direct interface. nftables is used for the firewalld primitives.

    def get_backend_by_name(self, name):
        for backend in self.all_backends():
            if backend.name == name:
                return backend
        raise FirewallError(errors.UNKNOWN_ERROR, "'%s' backend does not exist" % name)

    def get_backend_by_ipv(self, ipv):
        if self.nftables_enabled:
            return self.nftables_backend
        if ipv == "ipv4" and self.ip4tables_enabled:
            return self.ip4tables_backend
        elif ipv == "ipv6" and self.ip6tables_enabled:
            return self.ip6tables_backend
        elif ipv == "eb" and self.ebtables_enabled:
            return self.ebtables_backend
        raise FirewallError(
            errors.INVALID_IPV, "'%s' is not a valid backend or is unavailable" % ipv
        )

    def get_direct_backend_by_ipv(self, ipv):
        if ipv == "ipv4" and self.ip4tables_enabled:
            return self.ip4tables_backend
        elif ipv == "ipv6" and self.ip6tables_enabled:
            return self.ip6tables_backend
        elif ipv == "eb" and self.ebtables_enabled:
            return self.ebtables_backend
        raise FirewallError(
            errors.INVALID_IPV, "'%s' is not a valid backend or is unavailable" % ipv
        )

    def is_backend_enabled(self, name):
        if name == "ip4tables":
            return self.ip4tables_enabled
        elif name == "ip6tables":
            return self.ip6tables_enabled
        elif name == "ebtables":
            return self.ebtables_enabled
        elif name == "nftables":
            return self.nftables_enabled
        return False

    def is_ipv_enabled(self, ipv):
        if self.nftables_enabled:
            return True
        if ipv == "ipv4":
            return self.ip4tables_enabled
        elif ipv == "ipv6":
            return self.ip6tables_enabled
        elif ipv == "eb":
            return self.ebtables_enabled
        return False

    def enabled_backends(self):
        backends = []
        if self.nftables_enabled:
            backends.append(self.nftables_backend)
        else:
            if self.ip4tables_enabled:
                backends.append(self.ip4tables_backend)
            if self.ip6tables_enabled:
                backends.append(self.ip6tables_backend)
            if self.ebtables_enabled:
                backends.append(self.ebtables_backend)
        return backends

    def all_backends(self):
        backends = []
        if self.ip4tables_enabled:
            backends.append(self.ip4tables_backend)
        if self.ip6tables_enabled:
            backends.append(self.ip6tables_backend)
        if self.ebtables_enabled:
            backends.append(self.ebtables_backend)
        if self.nftables_enabled:
            backends.append(self.nftables_backend)
        return backends

    def apply_default_tables(self, use_transaction=None):
        if use_transaction is None:
            transaction = FirewallTransaction(self)
        else:
            transaction = use_transaction

        for backend in self.enabled_backends():
            transaction.add_rules(backend, backend.build_default_tables())

        if use_transaction is None:
            transaction.execute(True)

    def apply_default_rules(self, use_transaction=None):
        if use_transaction is None:
            transaction = FirewallTransaction(self)
        else:
            transaction = use_transaction

        for backend in self.enabled_backends():
            rules = backend.build_default_rules(self._log_denied)
            transaction.add_rules(backend, rules)

        if self.is_ipv_enabled("ipv6"):
            ipv6_backend = self.get_backend_by_ipv("ipv6")
            if "raw" in ipv6_backend.get_available_tables():
                if self.ipv6_rpfilter_enabled:
                    rules = ipv6_backend.build_rpfilter_rules(self._log_denied)
                    transaction.add_rules(ipv6_backend, rules)

        if self.is_ipv_enabled("ipv6") and self._rfc3964_ipv4:
            rules = ipv6_backend.build_rfc3964_ipv4_rules()
            transaction.add_rules(ipv6_backend, rules)

        if use_transaction is None:
            transaction.execute(True)

    def may_skip_flush_direct_backends(self):
        if self.nftables_enabled and not self.direct.has_runtime_configuration():
            return True

        return False

    def flush_direct_backends(self, use_transaction=None):
        if use_transaction is None:
            transaction = FirewallTransaction(self)
        else:
            transaction = use_transaction

        for backend in self.all_backends():
            if backend in self.enabled_backends():
                continue
            rules = backend.build_flush_rules()
            transaction.add_rules(backend, rules)

        if use_transaction is None:
            transaction.execute(True)

    def flush(self, use_transaction=None):
        if use_transaction is None:
            transaction = FirewallTransaction(self)
        else:
            transaction = use_transaction

        log.debug1("Flushing rule set")

        if not self.may_skip_flush_direct_backends():
            self.flush_direct_backends(use_transaction=transaction)

        for backend in self.enabled_backends():
            rules = backend.build_flush_rules()
            transaction.add_rules(backend, rules)

        if use_transaction is None:
            transaction.execute(True)

    def _set_policy_build_rules(self, backend, policy, policy_details=None):
        assert policy in ("ACCEPT", "DROP", "PANIC")
        if policy_details is None:
            dp = "ACCEPT" if policy == "ACCEPT" else "DROP"
            policy_details = {
                "INPUT": dp,
                "OUTPUT": dp,
                "FORWARD": dp,
            }
        return backend.build_set_policy_rules(policy, policy_details)

    def set_policy(self, policy, policy_details=None, use_transaction=None):
        if use_transaction is None:
            transaction = FirewallTransaction(self)
        else:
            transaction = use_transaction

        log.debug1(
            "Setting policy to '%s'%s",
            policy,
            f" (ReloadPolicy={firewalld_conf._unparse_reload_policy(policy_details)})"
            if policy == "DROP"
            else "",
        )

        for backend in self.enabled_backends():
            rules = self._set_policy_build_rules(backend, policy, policy_details)
            transaction.add_rules(backend, rules)

        if use_transaction is None:
            transaction.execute(True)

    # rule function used in handle_ functions

    def rule(self, backend_name, rule):
        if not rule:
            return ""

        backend = self.get_backend_by_name(backend_name)
        if not backend:
            raise FirewallError(
                errors.INVALID_IPV, "'%s' is not a valid backend" % backend_name
            )

        if not self.is_backend_enabled(backend_name):
            return ""

        return backend.set_rule(rule, self._log_denied)

    def rules(self, backend_name, rules):
        _rules = list(filter(None, rules))

        backend = self.get_backend_by_name(backend_name)
        if not backend:
            raise FirewallError(
                errors.INVALID_IPV, "'%s' is not a valid backend" % backend_name
            )

        if not self.is_backend_enabled(backend_name):
            return

        if (
            self._individual_calls
            or not backend.restore_command_exists
            or (
                backend_name == "ebtables"
                and not self.ebtables_backend.restore_noflush_option
            )
        ):
            for i, rule in enumerate(_rules):
                try:
                    backend.set_rule(rule, self._log_denied)
                except Exception as msg:
                    log.debug1(traceback.format_exc())
                    log.error(msg)
                    for rrule in reversed(_rules[:i]):
                        try:
                            backend.set_rule(
                                backend.reverse_rule(rrule), self._log_denied
                            )
                        except Exception:
                            # ignore errors here
                            pass
                    raise msg
        else:
            backend.set_rules(_rules, self._log_denied)

    # check functions

    def check_panic(self):
        if self._panic:
            raise FirewallError(errors.PANIC_MODE)

    def check_policy(self, policy):
        _policy = policy
        if _policy not in self.policy.get_policies():
            raise FirewallError(errors.INVALID_POLICY, _policy)
        return _policy

    def check_zone(self, zone):
        _zone = zone
        if not _zone or _zone == "":
            _zone = self.get_default_zone()
        if _zone not in self.zone.get_zones():
            raise FirewallError(errors.INVALID_ZONE, _zone)
        return _zone

    def check_interface(self, interface):
        if not functions.checkInterface(interface):
            raise FirewallError(errors.INVALID_INTERFACE, interface)

    def check_service(self, service):
        self.service.check_service(service)

    def check_port(self, port):
        if not functions.check_port(port):
            raise FirewallError(errors.INVALID_PORT, port)

    def check_tcpudp(self, protocol):
        if not protocol:
            raise FirewallError(errors.MISSING_PROTOCOL)
        if protocol not in ["tcp", "udp", "sctp", "dccp"]:
            raise FirewallError(
                errors.INVALID_PROTOCOL,
                "'%s' not in {'tcp'|'udp'|'sctp'|'dccp'}" % protocol,
            )

    def check_ip(self, ip):
        if not functions.checkIP(ip):
            raise FirewallError(errors.INVALID_ADDR, ip)

    def check_address(self, ipv, source):
        if ipv == "ipv4":
            if not functions.checkIPnMask(source):
                raise FirewallError(errors.INVALID_ADDR, source)
        elif ipv == "ipv6":
            if not functions.checkIP6nMask(source):
                raise FirewallError(errors.INVALID_ADDR, source)
        else:
            raise FirewallError(errors.INVALID_IPV, "'%s' not in {'ipv4'|'ipv6'}")

    def check_icmptype(self, icmp):
        self.icmptype.check_icmptype(icmp)

    def check_timeout(self, timeout):
        if not isinstance(timeout, int):
            raise TypeError("%s is %s, expected int" % (timeout, type(timeout)))
        if int(timeout) < 0:
            raise FirewallError(
                errors.INVALID_VALUE, "timeout '%d' is not positive number" % timeout
            )

    # RELOAD

    def reload(self, stop=False):
        # we're about to load the on-disk config, so verify it's sane.
        check_on_disk_config(self)

        _panic = self._panic
        _omit_native_ipset = self.ipset.omit_native_ipset()

        # must stash this. The value may change after _start()
        old_firewall_backend = self._firewall_backend
        flush_all = self._flush_all_on_reload

        if not flush_all:
            # save zone interfaces
            _zone_interfaces = {}
            for zone in self.zone.get_zones():
                _zone_interfaces[zone] = self.zone.get_zone(zone).interfaces
            # save direct config
            _direct_config = self.direct.get_runtime_config()
            _old_dz = self.get_default_zone()

        _ipset_objs = []
        for _name in self.ipset.get_ipsets():
            _ipset_objs.append(self.ipset.get_ipset(_name))

        if not _panic:
            reload_policy = firewalld_conf._parse_reload_policy(
                self._firewalld_conf.get("ReloadPolicy")
            )
            self.set_policy("DROP", policy_details=reload_policy)

        self.flush()
        self.cleanup()

        start_exception = None
        try:
            self._start(reload=True, complete_reload=stop)
        except Exception as e:
            # save the exception for later, but continue restoring interfaces,
            # etc. We'll re-raise it at the end.
            start_exception = e

        # destroy ipsets no longer in the permanent configuration
        if flush_all:
            for obj in _ipset_objs:
                if not self.ipset.query_ipset(obj.name):
                    # nftables sets are part of the normal firewall ruleset and
                    # thus do not need flushed here.
                    if self.ipset_enabled and not _omit_native_ipset:
                        self.ipset_backend.set_destroy(obj.name)

        if not flush_all:
            # handle interfaces in the default zone and move them to the new
            # default zone if it changed
            _new_dz = self.get_default_zone()
            if _new_dz != _old_dz:
                # if_new_dz has been introduced with the reload, we need to add it
                # https://github.com/firewalld/firewalld/issues/53
                if _new_dz not in _zone_interfaces:
                    _zone_interfaces[_new_dz] = {}
                # default zone changed. Move interfaces from old default zone to
                # the new one.
                for iface in _zone_interfaces[_old_dz]:
                    if iface in self._default_zone_interfaces:
                        # move only those that were added to default zone
                        # (not those that were added to specific zone same as
                        # default)
                        _zone_interfaces[_new_dz][iface] = _zone_interfaces[_old_dz][
                            iface
                        ]
                        del _zone_interfaces[_old_dz][iface]

            # add interfaces to zones again
            for zone in self.zone.get_zones():
                if zone in _zone_interfaces:
                    for interface_id in _zone_interfaces[zone]:
                        self.zone.change_zone_of_interface(zone, interface_id)

                    del _zone_interfaces[zone]
                else:
                    log.info1("New zone '%s'.", zone)
            if len(_zone_interfaces) > 0:
                for zone in list(_zone_interfaces.keys()):
                    log.info1("Lost zone '%s', zone interfaces dropped.", zone)
                    del _zone_interfaces[zone]
            del _zone_interfaces

            # restore runtime-only ipsets
            for obj in _ipset_objs:
                if self.ipset.query_ipset(obj.name):
                    for entry in obj.entries:
                        try:
                            self.ipset.add_entry(obj.name, entry)
                        except FirewallError as msg:
                            if msg.code != errors.ALREADY_ENABLED:
                                raise msg
                else:
                    self.ipset.add_ipset(obj)
                    self.ipset.apply_ipset(obj.name)

            # restore direct config
            self.direct.set_config(_direct_config)

        # Restore permanent interfaces from NetworkManager
        nm_bus_name = nm_get_bus_name()
        if nm_bus_name:
            for zone in self.zone.get_zones() + [""]:
                for interface in nm_get_interfaces_in_zone(zone):
                    self.zone.change_zone_of_interface(
                        zone, interface, sender=nm_bus_name
                    )

        self._panic = _panic
        if not self._panic:
            self.set_policy("ACCEPT")

        # If the FirewallBackend changed, then we must also cleanup the policy
        # for the old backend that was set to DROP above.
        if not self._panic and old_firewall_backend != self._firewall_backend:
            if old_firewall_backend == "nftables":
                for rule in self._set_policy_build_rules(
                    self.nftables_backend, "ACCEPT"
                ):
                    self.nftables_backend.set_rule(rule, self._log_denied)
            else:
                for rule in self._set_policy_build_rules(
                    self.ip4tables_backend, "ACCEPT"
                ):
                    self.ip4tables_backend.set_rule(rule, self._log_denied)
                if self.ip6tables_enabled:
                    for rule in self._set_policy_build_rules(
                        self.ip6tables_backend, "ACCEPT"
                    ):
                        self.ip6tables_backend.set_rule(rule, self._log_denied)

        if start_exception:
            self._state = "FAILED"
            raise start_exception
        else:
            self._state = "RUNNING"

    # STATE

    def get_state(self):
        return self._state

    # PANIC MODE

    def enable_panic_mode(self):
        if self._panic:
            raise FirewallError(errors.ALREADY_ENABLED, "panic mode already enabled")

        try:
            self.set_policy("PANIC")
        except Exception as msg:
            raise FirewallError(errors.COMMAND_FAILED, msg)
        self._panic = True

    def disable_panic_mode(self):
        if not self._panic:
            raise FirewallError(errors.NOT_ENABLED, "panic mode is not enabled")

        try:
            self.set_policy("ACCEPT")
        except Exception as msg:
            raise FirewallError(errors.COMMAND_FAILED, msg)
        self._panic = False

    def query_panic_mode(self):
        return self._panic

    # LOG DENIED

    def get_log_denied(self):
        return self._log_denied

    def set_log_denied(self, value):
        if value not in config.LOG_DENIED_VALUES:
            raise FirewallError(
                errors.INVALID_VALUE,
                "'%s', choose from '%s'"
                % (value, "','".join(config.LOG_DENIED_VALUES)),
            )

        if value != self.get_log_denied():
            self._log_denied = value
            self._firewalld_conf.set("LogDenied", value)
            self._firewalld_conf.write()
        else:
            raise FirewallError(errors.ALREADY_SET, value)

    # DEFAULT ZONE

    def get_default_zone(self):
        return self._default_zone

    def set_default_zone(self, zone):
        _zone = self.check_zone(zone)
        if _zone == self._default_zone:
            raise FirewallError(errors.ZONE_ALREADY_SET, _zone)

        self._firewalld_conf.set("DefaultZone", _zone)
        self._firewalld_conf.write()

    def combine_runtime_with_permanent_settings(self, permanent, runtime):
        combined = permanent.copy()

        for key, value in runtime.items():
            # omit empty entries
            if value or isinstance(value, bool) or isinstance(value, int):
                combined[key] = value
            # make sure to remove values that were in permanent, but no
            # longer in runtime.
            elif key in combined:
                del combined[key]

        return combined

    def get_added_and_removed_settings(self, old_settings, new_settings):
        # normalize rich rules, zones and policies use a different key
        for rich_key in ["rich_rules", "rules_str"]:
            if rich_key in new_settings:
                new_settings[rich_key] = [
                    str(Rich_Rule(rule_str=rule_str))
                    for rule_str in new_settings[rich_key]
                ]

        add_settings = {}
        remove_settings = {}
        for key in set(old_settings.keys()) | set(new_settings.keys()):
            if key in new_settings:
                if isinstance(new_settings[key], list):
                    old = set(old_settings[key] if key in old_settings else [])
                    add_settings[key] = list(set(new_settings[key]) - old)
                    remove_settings[key] = list((old ^ set(new_settings[key])) & old)
                # check for bool or int because dbus.Boolean is a subclass of
                # int (because bool can't be subclassed).
                elif isinstance(new_settings[key], bool) or isinstance(
                    new_settings[key], int
                ):
                    if not old_settings[key] and new_settings[key]:
                        add_settings[key] = True
                    elif old_settings[key] and not new_settings[key]:
                        remove_settings[key] = False
                else:
                    raise FirewallError(
                        errors.INVALID_SETTING,
                        "Unhandled setting type {} key {}".format(
                            type(new_settings[key]), key
                        ),
                    )

        return (add_settings, remove_settings)



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2012-2016 Red Hat, Inc.
#
# Authors:
# Thomas Woerner <twoerner@redhat.com>

from gi.repository import Gio, GLib


class Watcher:
    def __init__(self, callback, timeout):
        self._callback = callback
        self._timeout = timeout
        self._monitors = {}
        self._timeouts = {}
        self._blocked = []

    def add_watch_dir(self, directory):
        gfile = Gio.File.new_for_path(directory)
        self._monitors[directory] = gfile.monitor_directory(
            Gio.FileMonitorFlags.NONE, None
        )
        self._monitors[directory].connect("changed", self._file_changed_cb)

    def add_watch_file(self, filename):
        gfile = Gio.File.new_for_path(filename)
        self._monitors[filename] = gfile.monitor_file(Gio.FileMonitorFlags.NONE, None)
        self._monitors[filename].connect("changed", self._file_changed_cb)

    def get_watches(self):
        return self._monitors.keys()

    def has_watch(self, filename):
        return filename in self._monitors

    def remove_watch(self, filename):
        del self._monitors[filename]

    def block_source(self, filename):
        if filename not in self._blocked:
            self._blocked.append(filename)

    def unblock_source(self, filename):
        if filename in self._blocked:
            self._blocked.remove(filename)

    def clear_timeouts(self):
        for filename in list(self._timeouts.keys()):
            GLib.source_remove(self._timeouts[filename])
            del self._timeouts[filename]

    def _call_callback(self, filename):
        if filename not in self._blocked:
            self._callback(filename)
        del self._timeouts[filename]

    def _file_changed_cb(self, monitor, gio_file, gio_other_file, event):
        filename = gio_file.get_parse_name()
        if filename in self._blocked:
            if filename in self._timeouts:
                GLib.source_remove(self._timeouts[filename])
                del self._timeouts[filename]
            return

        if (
            event == Gio.FileMonitorEvent.CHANGED
            or event == Gio.FileMonitorEvent.CREATED
            or event == Gio.FileMonitorEvent.DELETED
            or event == Gio.FileMonitorEvent.ATTRIBUTE_CHANGED
        ):
            if filename in self._timeouts:
                GLib.source_remove(self._timeouts[filename])
                del self._timeouts[filename]
            self._timeouts[filename] = GLib.timeout_add_seconds(
                self._timeout, self._call_callback, filename
            )



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2010-2016 Red Hat, Inc.
#
# Authors:
# Thomas Woerner <twoerner@redhat.com>

"""modules backend"""

from firewall.core.prog import runProg
from firewall.core.logger import log
from firewall.config import COMMANDS


class modules:
    def __init__(self):
        self._load_command = COMMANDS["modprobe"]
        # Use rmmod instead of modprobe -r (RHBZ#1031102)
        self._unload_command = COMMANDS["rmmod"]

    def __repr__(self):
        return "%s" % (self.__class__)

    def loaded_modules(self):
        """get all loaded kernel modules and their dependencies"""
        mods = []
        deps = {}
        try:
            with open("/proc/modules", "r") as f:
                for line in f:
                    if not line:
                        break
                    line = line.strip()
                    splits = line.split()
                    mods.append(splits[0])
                    if splits[3] != "-":
                        deps[splits[0]] = splits[3].split(",")[:-1]
                    else:
                        deps[splits[0]] = []
        except FileNotFoundError:
            pass

        return mods, deps  # [loaded modules], {module:[dependants]}

    def load_module(self, module):
        log.debug2("%s: %s %s", self.__class__, self._load_command, module)
        return runProg(self._load_command, [module])

    def unload_module(self, module):
        log.debug2("%s: %s %s", self.__class__, self._unload_command, module)
        return runProg(self._unload_command, [module])

    def get_deps(self, module, deps, ret):
        """get all dependants of a module"""
        if module not in deps:
            return
        for mod in deps[module]:
            self.get_deps(mod, deps, ret)
            if mod not in ret:
                ret.append(mod)
        if module not in ret:
            ret.append(module)

    def get_firewall_modules(self):
        """get all loaded firewall-related modules"""
        mods = []
        (mods2, deps) = self.loaded_modules()

        self.get_deps("nf_conntrack", deps, mods)
        # these modules don't have dependants listed in /proc/modules
        for bad_bad_module in ["nf_conntrack_ipv4", "nf_conntrack_ipv6"]:
            if bad_bad_module in mods:
                # move them to end of list, so we'll remove them later
                mods.remove(bad_bad_module)
                mods.insert(-1, bad_bad_module)

        for mod in mods2:
            if (
                mod in ["ip_tables", "ip6_tables", "ebtables"]
                or mod.startswith("iptable_")
                or mod.startswith("ip6table_")
                or mod.startswith("nf_")
                or mod.startswith("xt_")
                or mod.startswith("ipt_")
                or mod.startswith("ip6t_")
            ):
                self.get_deps(mod, deps, mods)
        return mods

    def unload_firewall_modules(self):
        """unload all firewall-related modules"""
        for module in self.get_firewall_modules():
            (status, ret) = self.unload_module(module)
            if status != 0:
                log.debug1("Failed to unload module '%s': %s" % (module, ret))



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2017 Red Hat, Inc.
#
# Authors:
# Thomas Woerner <twoerner@redhat.com>

ICMP_TYPES = {
    "echo-reply": "0/0",
    "pong": "0/0",
    "network-unreachable": "3/0",
    "host-unreachable": "3/1",
    "protocol-unreachable": "3/2",
    "port-unreachable": "3/3",
    "fragmentation-needed": "3/4",
    "source-route-failed": "3/5",
    "network-unknown": "3/6",
    "host-unknown": "3/7",
    "network-prohibited": "3/9",
    "host-prohibited": "3/10",
    "TOS-network-unreachable": "3/11",
    "TOS-host-unreachable": "3/12",
    "communication-prohibited": "3/13",
    "host-precedence-violation": "3/14",
    "precedence-cutoff": "3/15",
    "source-quench": "4/0",
    "network-redirect": "5/0",
    "host-redirect": "5/1",
    "TOS-network-redirect": "5/2",
    "TOS-host-redirect": "5/3",
    "echo-request": "8/0",
    "ping": "8/0",
    "router-advertisement": "9/0",
    "router-solicitation": "10/0",
    "ttl-zero-during-transit": "11/0",
    "ttl-zero-during-reassembly": "11/1",
    "ip-header-bad": "12/0",
    "required-option-missing": "12/1",
    "timestamp-request": "13/0",
    "timestamp-reply": "14/0",
    "address-mask-request": "17/0",
    "address-mask-reply": "18/0",
}

ICMPV6_TYPES = {
    "no-route": "1/0",
    "communication-prohibited": "1/1",
    "address-unreachable": "1/3",
    "port-unreachable": "1/4",
    "packet-too-big": "2/0",
    "ttl-zero-during-transit": "3/0",
    "ttl-zero-during-reassembly": "3/1",
    "bad-header": "4/0",
    "unknown-header-type": "4/1",
    "unknown-option": "4/2",
    "echo-request": "128/0",
    "ping": "128/0",
    "echo-reply": "129/0",
    "pong": "129/0",
    "router-solicitation": "133/0",
    "router-advertisement": "134/0",
    "neighbour-solicitation": "135/0",
    "neigbour-solicitation": "135/0",
    "neighbour-advertisement": "136/0",
    "neigbour-advertisement": "136/0",
    "redirect": "137/0",
}


def check_icmp_name(_name):
    if _name in ICMP_TYPES:
        return True
    return False


def check_icmp_type(_type):
    if _type in ICMP_TYPES.values():
        return True
    return False


def check_icmpv6_name(_name):
    if _name in ICMPV6_TYPES:
        return True
    return False


def check_icmpv6_type(_type):
    if _type in ICMPV6_TYPES.values():
        return True
    return False



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2010-2016 Red Hat, Inc.
#
# Authors:
# Thomas Woerner <twoerner@redhat.com>

import dbus
import dbus.service

from firewall import config
from firewall.dbus_utils import (
    dbus_to_python,
    dbus_introspection_prepare_properties,
    dbus_introspection_add_properties,
)
from firewall.core.logger import log
from firewall.server.dbus import DbusServiceObject
from firewall.server.decorators import (
    handle_exceptions,
    dbus_handle_exceptions,
    dbus_service_method,
    dbus_polkit_require_auth,
)
from firewall import errors
from firewall.errors import FirewallError

############################################################################
#
# class FirewallDConfig
#
############################################################################


class FirewallDConfigService(DbusServiceObject):
    """FirewallD main class"""

    persistent = True
    """ Make FirewallD persistent. """
    default_polkit_auth_required = config.dbus.PK_ACTION_CONFIG
    """ Use PK_ACTION_INFO as a default """

    @handle_exceptions
    def __init__(self, parent, conf, service, item_id, *args, **kwargs):
        super(FirewallDConfigService, self).__init__(*args, **kwargs)
        self.parent = parent
        self.config = conf
        self.obj = service
        self.item_id = item_id
        self.busname = args[0]
        self.path = args[1]
        self._log_prefix = "config.service.%d" % self.item_id
        dbus_introspection_prepare_properties(
            self, config.dbus.DBUS_INTERFACE_CONFIG_SERVICE
        )

    @dbus_handle_exceptions
    def __del__(self):
        pass

    @dbus_handle_exceptions
    def unregister(self):
        self.remove_from_connection()

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    # P R O P E R T I E S

    @dbus_handle_exceptions
    def _get_property(self, property_name):
        if property_name == "name":
            return dbus.String(self.obj.name)
        elif property_name == "filename":
            return dbus.String(self.obj.filename)
        elif property_name == "path":
            return dbus.String(self.obj.path)
        elif property_name == "default":
            return dbus.Boolean(self.obj.default)
        elif property_name == "builtin":
            return dbus.Boolean(self.obj.builtin)
        else:
            raise dbus.exceptions.DBusException(
                "org.freedesktop.DBus.Error.InvalidArgs: "
                "Property '%s' does not exist" % property_name
            )

    @dbus_service_method(dbus.PROPERTIES_IFACE, in_signature="ss", out_signature="v")
    @dbus_handle_exceptions
    def Get(self, interface_name, property_name, sender=None):  # pylint: disable=W0613
        # get a property
        interface_name = dbus_to_python(interface_name, str)
        property_name = dbus_to_python(property_name, str)
        log.debug1(
            "%s.Get('%s', '%s')", self._log_prefix, interface_name, property_name
        )

        if interface_name != config.dbus.DBUS_INTERFACE_CONFIG_SERVICE:
            raise dbus.exceptions.DBusException(
                "org.freedesktop.DBus.Error.UnknownInterface: "
                "Interface '%s' does not exist" % interface_name
            )

        return self._get_property(property_name)

    @dbus_service_method(dbus.PROPERTIES_IFACE, in_signature="s", out_signature="a{sv}")
    @dbus_handle_exceptions
    def GetAll(self, interface_name, sender=None):  # pylint: disable=W0613
        interface_name = dbus_to_python(interface_name, str)
        log.debug1("%s.GetAll('%s')", self._log_prefix, interface_name)

        if interface_name != config.dbus.DBUS_INTERFACE_CONFIG_SERVICE:
            raise dbus.exceptions.DBusException(
                "org.freedesktop.DBus.Error.UnknownInterface: "
                "Interface '%s' does not exist" % interface_name
            )

        ret = {}
        for x in ["name", "filename", "path", "default", "builtin"]:
            ret[x] = self._get_property(x)
        return dbus.Dictionary(ret, signature="sv")

    @dbus_polkit_require_auth(config.dbus.PK_ACTION_CONFIG)
    @dbus_service_method(dbus.PROPERTIES_IFACE, in_signature="ssv")
    @dbus_handle_exceptions
    def Set(self, interface_name, property_name, new_value, sender=None):
        interface_name = dbus_to_python(interface_name, str)
        property_name = dbus_to_python(property_name, str)
        new_value = dbus_to_python(new_value)
        log.debug1(
            "%s.Set('%s', '%s', '%s')",
            self._log_prefix,
            interface_name,
            property_name,
            new_value,
        )
        self.parent.accessCheck(sender)

        if interface_name != config.dbus.DBUS_INTERFACE_CONFIG_SERVICE:
            raise dbus.exceptions.DBusException(
                "org.freedesktop.DBus.Error.UnknownInterface: "
                "Interface '%s' does not exist" % interface_name
            )

        raise dbus.exceptions.DBusException(
            "org.freedesktop.DBus.Error.PropertyReadOnly: "
            "Property '%s' is read-only" % property_name
        )

    @dbus.service.signal(dbus.PROPERTIES_IFACE, signature="sa{sv}as")
    def PropertiesChanged(
        self, interface_name, changed_properties, invalidated_properties
    ):
        interface_name = dbus_to_python(interface_name, str)
        changed_properties = dbus_to_python(changed_properties)
        invalidated_properties = dbus_to_python(invalidated_properties)
        log.debug1(
            "%s.PropertiesChanged('%s', '%s', '%s')",
            self._log_prefix,
            interface_name,
            changed_properties,
            invalidated_properties,
        )

    @dbus_polkit_require_auth(config.dbus.PK_ACTION_INFO)
    @dbus_service_method(dbus.INTROSPECTABLE_IFACE, out_signature="s")
    @dbus_handle_exceptions
    def Introspect(self, sender=None):  # pylint: disable=W0613
        log.debug2("%s.Introspect()", self._log_prefix)

        data = super(FirewallDConfigService, self).Introspect(
            self.path, self.busname.get_bus()
        )

        return dbus_introspection_add_properties(
            self, data, config.dbus.DBUS_INTERFACE_CONFIG_SERVICE
        )

    # S E T T I N G S

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_SERVICE,
        out_signature="(sssa(ss)asa{ss}asa(ss))",
    )
    @dbus_handle_exceptions
    def getSettings(self, sender=None):  # pylint: disable=W0613
        """get settings for service"""
        log.debug1("%s.getSettings()", self._log_prefix)
        return self.config.get_service_config(self.obj)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, out_signature="a{sv}"
    )
    @dbus_handle_exceptions
    def getSettings2(self, sender=None):
        """get settings for service"""
        log.debug1("%s.getSettings2()", self._log_prefix)
        return self.config.get_service_config_dict(self.obj)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_SERVICE,
        in_signature="(sssa(ss)asa{ss}asa(ss))",
    )
    @dbus_handle_exceptions
    def update(self, settings, sender=None):
        """update settings for service"""
        settings = dbus_to_python(settings)
        log.debug1("%s.update('...')", self._log_prefix)
        self.parent.accessCheck(sender)
        self.obj = self.config.set_service_config(self.obj, settings)
        self.Updated(self.obj.name)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="a{sv}"
    )
    @dbus_handle_exceptions
    def update2(self, settings, sender=None):
        settings = dbus_to_python(settings)
        log.debug1("%s.update2('...')", self._log_prefix)
        self.parent.accessCheck(sender)
        self.obj = self.config.set_service_config_dict(self.obj, settings)
        self.Updated(self.obj.name)

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE)
    @dbus_handle_exceptions
    def loadDefaults(self, sender=None):
        """load default settings for builtin service"""
        log.debug1("%s.loadDefaults()", self._log_prefix)
        self.parent.accessCheck(sender)
        self.obj = self.config.load_service_defaults(self.obj)
        self.Updated(self.obj.name)

    @dbus.service.signal(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, signature="s")
    @dbus_handle_exceptions
    def Updated(self, name):
        log.debug1("%s.Updated('%s')" % (self._log_prefix, name))

    # R E M O V E

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE)
    @dbus_handle_exceptions
    def remove(self, sender=None):
        """remove service"""
        log.debug1("%s.removeService()", self._log_prefix)
        self.parent.accessCheck(sender)
        self.config.remove_service(self.obj)
        self.parent.removeService(self.obj)

    @dbus.service.signal(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, signature="s")
    @dbus_handle_exceptions
    def Removed(self, name):
        log.debug1("%s.Removed('%s')" % (self._log_prefix, name))

    # R E N A M E

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="s")
    @dbus_handle_exceptions
    def rename(self, name, sender=None):
        """rename service"""
        name = dbus_to_python(name, str)
        log.debug1("%s.rename('%s')", self._log_prefix, name)
        self.parent.accessCheck(sender)
        self.obj = self.config.rename_service(self.obj, name)
        self.Renamed(name)

    @dbus.service.signal(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, signature="s")
    @dbus_handle_exceptions
    def Renamed(self, name):
        log.debug1("%s.Renamed('%s')" % (self._log_prefix, name))

    # version

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, out_signature="s")
    @dbus_handle_exceptions
    def getVersion(self, sender=None):  # pylint: disable=W0613
        log.debug1("%s.getVersion()", self._log_prefix)
        return self.getSettings()[0]

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="s")
    @dbus_handle_exceptions
    def setVersion(self, version, sender=None):
        version = dbus_to_python(version, str)
        log.debug1("%s.setVersion('%s')", self._log_prefix, version)
        self.parent.accessCheck(sender)
        settings = list(self.getSettings())
        settings[0] = version
        self.update(settings)

    # short

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, out_signature="s")
    @dbus_handle_exceptions
    def getShort(self, sender=None):  # pylint: disable=W0613
        log.debug1("%s.getShort()", self._log_prefix)
        return self.getSettings()[1]

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="s")
    @dbus_handle_exceptions
    def setShort(self, short, sender=None):
        short = dbus_to_python(short, str)
        log.debug1("%s.setShort('%s')", self._log_prefix, short)
        self.parent.accessCheck(sender)
        settings = list(self.getSettings())
        settings[1] = short
        self.update(settings)

    # description

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, out_signature="s")
    @dbus_handle_exceptions
    def getDescription(self, sender=None):  # pylint: disable=W0613
        log.debug1("%s.getDescription()", self._log_prefix)
        return self.getSettings()[2]

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="s")
    @dbus_handle_exceptions
    def setDescription(self, description, sender=None):
        description = dbus_to_python(description, str)
        log.debug1("%s.setDescription('%s')", self._log_prefix, description)
        self.parent.accessCheck(sender)
        settings = list(self.getSettings())
        settings[2] = description
        self.update(settings)

    # port

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, out_signature="a(ss)"
    )
    @dbus_handle_exceptions
    def getPorts(self, sender=None):  # pylint: disable=W0613
        log.debug1("%s.getPorts()", self._log_prefix)
        return self.getSettings()[3]

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="a(ss)"
    )
    @dbus_handle_exceptions
    def setPorts(self, ports, sender=None):
        _ports = []
        # convert embedded lists to tuples
        for port in dbus_to_python(ports, list):
            if isinstance(port, list):
                _ports.append(tuple(port))
            else:
                _ports.append(port)
        ports = _ports
        log.debug1(
            "%s.setPorts('[%s]')",
            self._log_prefix,
            ",".join("('%s, '%s')" % (port[0], port[1]) for port in ports),
        )
        self.parent.accessCheck(sender)
        settings = list(self.getSettings())
        settings[3] = ports
        self.update(settings)

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="ss")
    @dbus_handle_exceptions
    def addPort(self, port, protocol, sender=None):
        port = dbus_to_python(port, str)
        protocol = dbus_to_python(protocol, str)
        log.debug1("%s.addPort('%s', '%s')", self._log_prefix, port, protocol)
        self.parent.accessCheck(sender)
        settings = list(self.getSettings())
        if (port, protocol) in settings[3]:
            raise FirewallError(errors.ALREADY_ENABLED, "%s:%s" % (port, protocol))
        settings[3].append((port, protocol))
        self.update(settings)

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="ss")
    @dbus_handle_exceptions
    def removePort(self, port, protocol, sender=None):
        port = dbus_to_python(port, str)
        protocol = dbus_to_python(protocol, str)
        log.debug1("%s.removePort('%s', '%s')", self._log_prefix, port, protocol)
        self.parent.accessCheck(sender)
        settings = list(self.getSettings())
        if (port, protocol) not in settings[3]:
            raise FirewallError(errors.NOT_ENABLED, "%s:%s" % (port, protocol))
        settings[3].remove((port, protocol))
        self.update(settings)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="ss", out_signature="b"
    )
    @dbus_handle_exceptions
    def queryPort(self, port, protocol, sender=None):  # pylint: disable=W0613
        port = dbus_to_python(port, str)
        protocol = dbus_to_python(protocol, str)
        log.debug1("%s.queryPort('%s', '%s')", self._log_prefix, port, protocol)
        return (port, protocol) in self.getSettings()[3]

    # protocol

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, out_signature="as")
    @dbus_handle_exceptions
    def getProtocols(self, sender=None):  # pylint: disable=W0613
        log.debug1("%s.getProtocols()", self._log_prefix)
        return self.getSettings()[6]

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="as")
    @dbus_handle_exceptions
    def setProtocols(self, protocols, sender=None):
        protocols = dbus_to_python(protocols, list)
        log.debug1("%s.setProtocols('[%s]')", self._log_prefix, ",".join(protocols))
        self.parent.accessCheck(sender)
        settings = list(self.getSettings())
        settings[6] = protocols
        self.update(settings)

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="s")
    @dbus_handle_exceptions
    def addProtocol(self, protocol, sender=None):
        protocol = dbus_to_python(protocol, str)
        log.debug1("%s.addProtocol('%s')", self._log_prefix, protocol)
        self.parent.accessCheck(sender)
        settings = list(self.getSettings())
        if protocol in settings[6]:
            raise FirewallError(errors.ALREADY_ENABLED, protocol)
        settings[6].append(protocol)
        self.update(settings)

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="s")
    @dbus_handle_exceptions
    def removeProtocol(self, protocol, sender=None):
        protocol = dbus_to_python(protocol, str)
        log.debug1("%s.removeProtocol('%s')", self._log_prefix, protocol)
        self.parent.accessCheck(sender)
        settings = list(self.getSettings())
        if protocol not in settings[6]:
            raise FirewallError(errors.NOT_ENABLED, protocol)
        settings[6].remove(protocol)
        self.update(settings)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="s", out_signature="b"
    )
    @dbus_handle_exceptions
    def queryProtocol(self, protocol, sender=None):  # pylint: disable=W0613
        protocol = dbus_to_python(protocol, str)
        log.debug1("%s.queryProtocol(%s')", self._log_prefix, protocol)
        return protocol in self.getSettings()[6]

    # source port

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, out_signature="a(ss)"
    )
    @dbus_handle_exceptions
    def getSourcePorts(self, sender=None):  # pylint: disable=W0613
        log.debug1("%s.getSourcePorts()", self._log_prefix)
        return self.getSettings()[7]

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="a(ss)"
    )
    @dbus_handle_exceptions
    def setSourcePorts(self, ports, sender=None):
        _ports = []
        # convert embedded lists to tuples
        for port in dbus_to_python(ports, list):
            if isinstance(port, list):
                _ports.append(tuple(port))
            else:
                _ports.append(port)
        ports = _ports
        log.debug1(
            "%s.setSourcePorts('[%s]')",
            self._log_prefix,
            ",".join("('%s, '%s')" % (port[0], port[1]) for port in ports),
        )
        self.parent.accessCheck(sender)
        settings = list(self.getSettings())
        settings[7] = ports
        self.update(settings)

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="ss")
    @dbus_handle_exceptions
    def addSourcePort(self, port, protocol, sender=None):
        port = dbus_to_python(port, str)
        protocol = dbus_to_python(protocol, str)
        log.debug1("%s.addSourcePort('%s', '%s')", self._log_prefix, port, protocol)
        self.parent.accessCheck(sender)
        settings = list(self.getSettings())
        if (port, protocol) in settings[7]:
            raise FirewallError(errors.ALREADY_ENABLED, "%s:%s" % (port, protocol))
        settings[7].append((port, protocol))
        self.update(settings)

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="ss")
    @dbus_handle_exceptions
    def removeSourcePort(self, port, protocol, sender=None):
        port = dbus_to_python(port, str)
        protocol = dbus_to_python(protocol, str)
        log.debug1("%s.removeSourcePort('%s', '%s')", self._log_prefix, port, protocol)
        self.parent.accessCheck(sender)
        settings = list(self.getSettings())
        if (port, protocol) not in settings[7]:
            raise FirewallError(errors.NOT_ENABLED, "%s:%s" % (port, protocol))
        settings[7].remove((port, protocol))
        self.update(settings)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="ss", out_signature="b"
    )
    @dbus_handle_exceptions
    def querySourcePort(self, port, protocol, sender=None):  # pylint: disable=W0613
        port = dbus_to_python(port, str)
        protocol = dbus_to_python(protocol, str)
        log.debug1("%s.querySourcePort('%s', '%s')", self._log_prefix, port, protocol)
        return (port, protocol) in self.getSettings()[7]

    # module

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, out_signature="as")
    @dbus_handle_exceptions
    def getModules(self, sender=None):  # pylint: disable=W0613
        log.debug1("%s.getModules()", self._log_prefix)
        return self.getSettings()[4]

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="as")
    @dbus_handle_exceptions
    def setModules(self, modules, sender=None):
        modules = dbus_to_python(modules, list)
        _modules = []
        for module in modules:
            if module.startswith("nf_conntrack_"):
                module = module.replace("nf_conntrack_", "")
                if "_" in module:
                    module = module.replace("_", "-")
            _modules.append(module)
        modules = _modules
        log.debug1("%s.setModules('[%s]')", self._log_prefix, ",".join(modules))
        self.parent.accessCheck(sender)
        settings = list(self.getSettings())
        settings[4] = modules
        self.update(settings)

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="s")
    @dbus_handle_exceptions
    def addModule(self, module, sender=None):
        module = dbus_to_python(module, str)
        if module.startswith("nf_conntrack_"):
            module = module.replace("nf_conntrack_", "")
            if "_" in module:
                module = module.replace("_", "-")
        log.debug1("%s.addModule('%s')", self._log_prefix, module)
        self.parent.accessCheck(sender)
        settings = list(self.getSettings())
        if module in settings[4]:
            raise FirewallError(errors.ALREADY_ENABLED, module)
        settings[4].append(module)
        self.update(settings)

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="s")
    @dbus_handle_exceptions
    def removeModule(self, module, sender=None):
        module = dbus_to_python(module, str)
        if module.startswith("nf_conntrack_"):
            module = module.replace("nf_conntrack_", "")
            if "_" in module:
                module = module.replace("_", "-")
        log.debug1("%s.removeModule('%s')", self._log_prefix, module)
        self.parent.accessCheck(sender)
        settings = list(self.getSettings())
        if module not in settings[4]:
            raise FirewallError(errors.NOT_ENABLED, module)
        settings[4].remove(module)
        self.update(settings)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="s", out_signature="b"
    )
    @dbus_handle_exceptions
    def queryModule(self, module, sender=None):  # pylint: disable=W0613
        module = dbus_to_python(module, str)
        if module.startswith("nf_conntrack_"):
            module = module.replace("nf_conntrack_", "")
            if "_" in module:
                module = module.replace("_", "-")
        log.debug1("%s.queryModule('%s')", self._log_prefix, module)
        return module in self.getSettings()[4]

    # destination

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, out_signature="a{ss}"
    )
    @dbus_handle_exceptions
    def getDestinations(self, sender=None):  # pylint: disable=W0613
        log.debug1("%s.getDestinations()", self._log_prefix)
        return self.getSettings()[5]

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="a{ss}"
    )
    @dbus_handle_exceptions
    def setDestinations(self, destinations, sender=None):
        destinations = dbus_to_python(destinations, dict)
        log.debug1(
            "%s.setDestinations({ipv4:'%s', ipv6:'%s'})",
            self._log_prefix,
            destinations.get("ipv4"),
            destinations.get("ipv6"),
        )
        self.parent.accessCheck(sender)
        settings = list(self.getSettings())
        settings[5] = destinations
        self.update(settings)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="s", out_signature="s"
    )
    @dbus_handle_exceptions
    def getDestination(self, family, sender=None):
        family = dbus_to_python(family, str)
        log.debug1("%s.getDestination('%s')", self._log_prefix, family)
        self.parent.accessCheck(sender)
        settings = list(self.getSettings())
        if family not in settings[5]:
            raise FirewallError(errors.NOT_ENABLED, family)
        return settings[5][family]

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="ss")
    @dbus_handle_exceptions
    def setDestination(self, family, address, sender=None):
        family = dbus_to_python(family, str)
        address = dbus_to_python(address, str)
        log.debug1("%s.setDestination('%s', '%s')", self._log_prefix, family, address)
        self.parent.accessCheck(sender)
        settings = list(self.getSettings())
        if family in settings[5] and settings[5][family] == address:
            raise FirewallError(
                errors.ALREADY_ENABLED, "'%s': '%s'" % (family, address)
            )
        settings[5][family] = address
        self.update(settings)

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="s")
    @dbus_handle_exceptions
    def removeDestination(self, family, sender=None):
        family = dbus_to_python(family, str)
        log.debug1("%s.removeDestination('%s')", self._log_prefix, family)
        self.parent.accessCheck(sender)
        settings = list(self.getSettings())
        if family not in settings[5]:
            raise FirewallError(errors.NOT_ENABLED, family)
        del settings[5][family]
        self.update(settings)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="ss", out_signature="b"
    )
    @dbus_handle_exceptions
    def queryDestination(self, family, address, sender=None):  # pylint: disable=W0613
        family = dbus_to_python(family, str)
        address = dbus_to_python(address, str)
        log.debug1("%s.queryDestination('%s', '%s')", self._log_prefix, family, address)
        settings = self.getSettings()
        return family in settings[5] and address == settings[5][family]

    # includes

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, out_signature="as")
    @dbus_handle_exceptions
    def getIncludes(self, sender=None):
        log.debug1("%s.getIncludes()", self._log_prefix)
        self.parent.accessCheck(sender)
        settings = self.config.get_service_config_dict(self.obj)
        return settings["includes"] if "includes" in settings else []

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="as")
    @dbus_handle_exceptions
    def setIncludes(self, includes, sender=None):
        includes = dbus_to_python(includes, list)
        log.debug1("%s.setIncludes('%s')", self._log_prefix, includes)
        self.parent.accessCheck(sender)
        settings = {"includes": includes[:]}
        self.obj = self.config.set_service_config_dict(self.obj, settings)
        self.Updated(self.obj.name)

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="s")
    @dbus_handle_exceptions
    def addInclude(self, include, sender=None):
        include = dbus_to_python(include, str)
        log.debug1("%s.addInclude('%s')", self._log_prefix, include)
        self.parent.accessCheck(sender)
        settings = self.config.get_service_config_dict(self.obj)
        settings.setdefault("includes", []).append(include)
        self.obj = self.config.set_service_config_dict(self.obj, settings)
        self.Updated(self.obj.name)

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="s")
    @dbus_handle_exceptions
    def removeInclude(self, include, sender=None):
        include = dbus_to_python(include, str)
        log.debug1("%s.removeInclude('%s')", self._log_prefix, include)
        self.parent.accessCheck(sender)
        settings = self.config.get_service_config_dict(self.obj)
        settings["includes"].remove(include)
        self.obj = self.config.set_service_config_dict(self.obj, settings)
        self.Updated(self.obj.name)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_SERVICE, in_signature="s", out_signature="b"
    )
    @dbus_handle_exceptions
    def queryInclude(self, include, sender=None):
        include = dbus_to_python(include, str)
        log.debug1("%s.queryInclude('%s')", self._log_prefix, include)
        settings = self.config.get_service_config_dict(self.obj)
        return include in settings["includes"] if "includes" in settings else False
